#include <string.h>
#include "x++.hpp"
#include "engineaudio.hpp"
#include "audio.h"


namespace {
  
  const int MAX_SOUNDS = 16;

  CMusicModule Music;
  CSoundEffect Sounds[MAX_SOUNDS];

  int CurrentSound;
};

unsigned long AudioID_Count = 1;

////////////////////////////////////////////////////////////////////////////////

bool PlayMusic(const char* name, FILE* file, long fsize)
{
  // destroy the old song
  Music.Destroy();

  // if no song is specified, just return
  if (name == NULL || strlen(name) == 0) {
    return true;
  }

  Music.SetFile(file);
  Music.SetFileSize(fsize);
  // load the music file
  if(!Music.Load(name))
  {
    if (file != NULL)
      fclose(file);
    return false;
  }

  // play it
  Music.Play();
  return true;
}

////////////////////////////////////////////////////////////////////////////////

bool PlaySound(const char* name, FILE* file, long fsize)
{

  // PlaySound("") stops all sounds
  if (name == NULL || strlen(name) == 0) {
    for (int i = 0; i < MAX_SOUNDS; i++) {
      Sounds[i].Destroy();
    }

    CurrentSound = 0;
    return true;
  }

  // destroy the currently playing sound
    Sounds[CurrentSound].Destroy();

  Sounds[CurrentSound].SetFile(file);
  Sounds[CurrentSound].SetFileSize(fsize);
  // load the new sound
  if (!Sounds[CurrentSound].Load(name))
  {
    fclose(file);
    return false;
  }

  // begin playing it
  Sounds[CurrentSound].Play();

  CurrentSound = (CurrentSound + 1) % MAX_SOUNDS;
  return true;
}

////////////////////////////////////////////////////////////////////////////////

bool PlaySoundAbsolute(const char* name, int volume, FILE* file, long fsize)
{
  // PlaySound("") stops all sounds
  if (name == NULL || strlen(name) == 0) {
    for (int i = 0; i < MAX_SOUNDS; i++) {
      Sounds[i].Destroy();
    }

    CurrentSound = 0;
    return true;
  }

  // destroy the currently playing sound
    Sounds[CurrentSound].Destroy();

  Sounds[CurrentSound].SetFile(file);
  Sounds[CurrentSound].SetFileSize(fsize);
  // load the new sound
  if (!Sounds[CurrentSound].Load(name))
  {
    fclose(file);
    return false;
  }

  // begin playing it
  Sounds[CurrentSound].PlayAbsolute(volume);

  CurrentSound = (CurrentSound + 1) % MAX_SOUNDS;
  return true;
}

////////////////////////////////////////////////////////////////////////////////

int GetSoundVolume()
{
  return GetGlobalSoundVolume();
}

////////////////////////////////////////////////////////////////////////////////

void SetSoundVolume(int volume)
{
  SetGlobalSoundVolume(volume);
  Music.SetVolume(GetMusicVolume());
}

////////////////////////////////////////////////////////////////////////////////

int GetMusicVolume()
{
  return GetGlobalMusicVolume();
}

////////////////////////////////////////////////////////////////////////////////

void SetMusicVolume(int volume)
{
  Music.SetVolume(volume);
  SetGlobalMusicVolume(volume);
}

////////////////////////////////////////////////////////////////////////////////

unsigned long AudioOpenFileCallback(char* filename)
{
  if (strcmp_ci(filename + strlen(filename) - 4, ".wav") == 0 ||
      strcmp_ci(filename + strlen(filename) - 4, ".raw") == 0 )
  {
    FILE* file = Sounds[CurrentSound].GetFile();
    if (file == NULL)
      return 0;

    // let's assume the current sound would the sound we want
    Sounds[CurrentSound].SetID(AudioID_Count);
    Sounds[CurrentSound].SetFileStartPosition(ftell(file));
  }
  else
  {
    FILE* file = Music.GetFile();
    if (file == NULL)
      return 0;

    Music.SetID(AudioID_Count);
    Music.SetFileStartPosition(ftell(file));
  }


  AudioID_Count++;
  return AudioID_Count - 1;
}

////////////////////////////////////////////////////////////////////////////////

void AudioCloseFileCallback(unsigned long id)
{
  if (id == Music.GetID())
  {
    fclose(Music.GetFile());
  }
  else
  {
    for(int i=0; i<MAX_SOUNDS; i++)
      if (id == Sounds[i].GetID())
        fclose(Sounds[i].GetFile());
  }
}

////////////////////////////////////////////////////////////////////////////////

long AudioReadFileCallback(void* buffer, long size, unsigned long id)
{
  long position;
  long bytes_copied;
  FILE* file;

  if (id == Music.GetID())
  {
    file = Music.GetFile();
    position = ftell(file);
    bytes_copied = size;

    // let's trim it up a bit if it's too big
    if (bytes_copied + position > Music.GetFileStartPosition() + Music.GetFileSize())
    {
      long bytes_deducted = (bytes_copied + position) - (Music.GetFileStartPosition() + Music.GetFileSize());
      bytes_copied -= bytes_deducted;
    }
  }
  else
  {
    for(int i=0; i<MAX_SOUNDS; i++)
      if (id == Sounds[i].GetID())
      {
        file = Sounds[i].GetFile();
        position = ftell(file);
        bytes_copied = size;

        // let's trim it up a bit if it's too big
        if (bytes_copied + position > Sounds[i].GetFileStartPosition() + Sounds[i].GetFileSize())
        {
          long bytes_deducted = (bytes_copied + position) - (Music.GetFileStartPosition() + Sounds[i].GetFileSize());
          bytes_copied -= bytes_deducted;
        }
      }
  }

  fread(buffer, 1, bytes_copied, file);
  return bytes_copied;
}

////////////////////////////////////////////////////////////////////////////////

void AudioSeekFileCallback(unsigned long id, long offset, signed char mode)
{
  if (id == Music.GetID())
  {
    FILE* file = Music.GetFile();
    long pos;

    if (mode == SEEK_CUR)
    {
      fseek(file, offset, SEEK_CUR);
    }
    else if (mode == SEEK_SET)
    {
      pos = Music.GetFileStartPosition();
      pos += offset;

      fseek(file, pos, SEEK_SET);
    }
    else if (mode == SEEK_END)
    {
      pos = Music.GetFileStartPosition();
      pos += Music.GetFileSize();
      pos += offset;

      fseek(file, pos, SEEK_SET);
    }
  }
  else
  {
    for(int i=0; i<MAX_SOUNDS; i++)
      if (id == Sounds[i].GetID())
      {
        FILE* file = Sounds[i].GetFile();
        long pos;

        if (mode == SEEK_CUR)
        {
          fseek(file, offset, SEEK_CUR);
        }
        else if (mode == SEEK_SET)
        {
          pos = Sounds[i].GetFileStartPosition();
          pos += offset;

          fseek(file, pos, SEEK_SET);
        }
        else if (mode == SEEK_END)
        {
          pos = Sounds[i].GetFileStartPosition();
          pos += Sounds[i].GetFileSize();
          pos += offset;

          fseek(file, pos, SEEK_SET);
        }
      }
  }
}

////////////////////////////////////////////////////////////////////////////////

long AudioTellFileCallback(unsigned long id)
{
  long position;

  if (id == Music.GetID())
  {
    position = ftell(Music.GetFile());
    position -= Music.GetFileStartPosition();
  }
  else
  {
    for(int i=0; i<MAX_SOUNDS; i++)
      if (id == Sounds[i].GetID())
      {
        position = ftell(Sounds[i].GetFile());
        position -= Music.GetFileStartPosition();
      }
  }

  return position;
}