#include <windows.h>
#include <mmsystem.h>
#include <stdio.h>
#include <stdlib.h>
#include "internal.h"
#include "fmod.h"

#define ENABLE_SOUNDEFFECTS
#define ENABLE_MUSICMODULES
//#define ENABLE_DIRECTMUSIC

#include "audio.h"


static HWND  SphereWindow;
static bool  Stereo;
static dword BaseMode;
static bool  AudioEnabled;
static int   MusicVolume;

////////////////////////////////////////////////////////////////////////////////

bool InitAudio(HWND window, SPHERECONFIG* config)
{
  int output_driver;
  int mixer;

  SphereWindow = window;

  // set output type
  output_driver = FSOUND_OUTPUT_NOSOUND;
  switch (config->output_driver)
  {
    case 0: output_driver = FSOUND_OUTPUT_NOSOUND; break;
    case 1: output_driver = FSOUND_OUTPUT_WINMM;   break;
    case 2: output_driver = FSOUND_OUTPUT_DSOUND;  break;
  }

retry:

  FSOUND_SetOutput(output_driver);

  // set mixer
  mixer = FSOUND_MIXER_AUTODETECT;
  switch (config->mixer)
  {
    case 0: mixer = FSOUND_MIXER_AUTODETECT; break;
    case 1: mixer = FSOUND_MIXER_QUALITY_AUTODETECT; break;
    case 2: mixer = FSOUND_MIXER_BLENDMODE; break;
    case 3: mixer = FSOUND_MIXER_MMXP5; break;
    case 4: mixer = FSOUND_MIXER_MMXP6; break;
    case 5: mixer = FSOUND_MIXER_QUALITY_FPU; break;
    case 6: mixer = FSOUND_MIXER_QUALITY_MMXP5; break;
    case 7: mixer = FSOUND_MIXER_QUALITY_MMXP6; break;
  }
  FSOUND_SetMixer(mixer);

  // initialize audio
  if (!FSOUND_Init(config->bit_rate, 64, 0))
  {
    // try reinitialization with no sound
    if (output_driver != FSOUND_OUTPUT_NOSOUND)
    {
      output_driver = FSOUND_OUTPUT_NOSOUND;
      goto retry;  // I know...  :)
    }

    return false;
  }

  // set up master music and sound volumes
  FSOUND_SetSFXMasterVolume(config->sound_volume);
  MusicVolume = config->music_volume;

  // calculate modes for MP3 streaming
  Stereo = config->stereo ? true : false;
  BaseMode = FSOUND_2D;
  BaseMode |= (config->stereo ? FSOUND_STEREO : FSOUND_MONO);
  BaseMode |= (config->bit_rate > 11025 ? FSOUND_16BITS : FSOUND_8BITS);


  FSOUND_File_SetCallbacks(AudioOpenFileCallback, AudioCloseFileCallback,
                           AudioReadFileCallback, AudioSeekFileCallback,
                           AudioTellFileCallback);
  AudioEnabled = (output_driver != FSOUND_OUTPUT_NOSOUND);

  #ifdef ENABLE_DIRECTMUSIC
    if (CoInitialize(NULL) < 0)
      return false;
  #else
    mciSendString("set time format milliseconds", NULL, NULL, NULL);
  #endif
  return true;
}

////////////////////////////////////////////////////////////////////////////////

bool CloseAudio(void)
{
  FSOUND_Close();
  #ifdef ENABLE_DIRECTMUSIC
    CoUninitialize();
  #endif
  return true;
}

////////////////////////////////////////////////////////////////////////////////

CSoundEffect::CSoundEffect()
: m_Loaded(false)
{
}

////////////////////////////////////////////////////////////////////////////////

CSoundEffect::~CSoundEffect()
{
}

////////////////////////////////////////////////////////////////////////////////

bool
CSoundEffect::Load(const char* filename)
{
#ifdef ENABLE_SOUNDEFFECTS
  if (stricmp(filename + strlen(filename) - 4, ".wav") == 0)
    sample = FSOUND_Sample_LoadWav(FSOUND_FREE, (char*) filename, BaseMode | FSOUND_LOOP_OFF);
  else if (stricmp(filename + strlen(filename) - 4, ".raw") == 0)
    sample = FSOUND_Sample_LoadRaw(FSOUND_FREE, (char*) filename, BaseMode | FSOUND_LOOP_OFF);
  else
    return false;

  if (sample == NULL)
    return false;

  m_Loaded = true;
  return true;

#else

  return true;

#endif
}

////////////////////////////////////////////////////////////////////////////////

void
CSoundEffect::Destroy()
{
#ifdef ENABLE_SOUNDEFFECTS

  Stop();
  if (m_Loaded)
    FSOUND_Sample_Free(sample);

#else

  fclose(fpointer);

#endif
}

////////////////////////////////////////////////////////////////////////////////

void
CSoundEffect::Play()
{
#ifdef ENABLE_SOUNDEFFECTS

  if (!AudioEnabled)
    return;

  m_Channel = FSOUND_PlaySound(FSOUND_FREE, sample);

#endif
}

////////////////////////////////////////////////////////////////////////////////

void
CSoundEffect::PlayAbsolute(int volume)
{
#ifdef ENABLE_SOUNDEFFECTS

  if (!AudioEnabled)
    return;

  m_Channel = FSOUND_PlaySound(FSOUND_FREE, sample);
  FSOUND_SetVolumeAbsolute(m_Channel, volume);

#endif
}

////////////////////////////////////////////////////////////////////////////////

void
CSoundEffect::Stop()
{
#ifdef ENABLE_SOUNDEFFECTS

  if (!AudioEnabled)
    return;

  if (!m_Loaded)
    return;

  if (FSOUND_IsPlaying(m_Channel))
    FSOUND_Sample_Free(sample);

#endif
}

////////////////////////////////////////////////////////////////////////////////

unsigned long 
CSoundEffect::GetID()
{
  return fhandle;
}

////////////////////////////////////////////////////////////////////////////////

void 
CSoundEffect::SetID(unsigned long id)
{
  fhandle = id;
}

////////////////////////////////////////////////////////////////////////////////

FILE* 
CSoundEffect::GetFile()
{
  return fpointer;
}

////////////////////////////////////////////////////////////////////////////////

void 
CSoundEffect::SetFile(FILE* file)
{
  fpointer = file;
}

////////////////////////////////////////////////////////////////////////////////

long
CSoundEffect::GetFileSize()
{
  return m_Size;
}

////////////////////////////////////////////////////////////////////////////////

void
CSoundEffect::SetFileSize(long size)
{
  m_Size = size;
}

////////////////////////////////////////////////////////////////////////////////

long
CSoundEffect::GetFileStartPosition()
{
  return m_FileStartPosition;
}

////////////////////////////////////////////////////////////////////////////////

void
CSoundEffect::SetFileStartPosition(long start)
{
  m_FileStartPosition = start;
}

////////////////////////////////////////////////////////////////////////////////

CMusicModule::CMusicModule()
: m_Loaded(false)
, m_IsPlaying(false)
{
}

////////////////////////////////////////////////////////////////////////////////

CMusicModule::~CMusicModule()
{
}

////////////////////////////////////////////////////////////////////////////////

bool
CMusicModule::Load(const char* filename)
{
#ifdef ENABLE_MUSICMODULES

  // If file is a MIDI
  if (stricmp(filename + strlen(filename) - 4, ".mid") == 0 ||
      stricmp(filename + strlen(filename) - 5, ".midi") == 0)
  {
    #ifdef ENABLE_DIRECTMUSIC
      if (FAILED(CoCreateInstance(CLSID_DirectMusicPerformance, NULL, CLSCTX_INPROC, IID_IDirectMusicPerformance2, (void**)&directmusicperformance)))
        return false;

      if (FAILED(CoCreateInstance(CLSID_DirectMusicLoader, NULL, CLSCTX_INPROC, IID_IDirectMusicLoader, (void**)&directmusicloader)))
      {
        directmusicperformance->Release();
        return false;
      }

      DMUS_OBJECTDESC ObjDesc;
      WCHAR wcFilename[MAX_PATH];

      ObjDesc.dwSize = sizeof(DMUS_OBJECTDESC);
      ObjDesc.dwValidData = DMUS_OBJ_CLASS | DMUS_OBJ_FILENAME | DMUS_OBJ_FULLPATH;
      ObjDesc.guidClass = CLSID_DirectMusicSegment;
      MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, filename, -1, wcFilename, MAX_PATH);
      wcscpy(ObjDesc.wszFileName, wcFilename);

      if (FAILED(directmusicloader->GetObject(&ObjDesc, IID_IDirectMusicSegment2, (void**)&directmusicsegment)))
      {
        directmusicloader->Release();
        directmusicperformance->Release();
        return false;
      }

      directmusic = NULL;
      directmusicperformance->Init(&directmusic, NULL, NULL);
      directmusicperformance->AddPort(NULL);
      directmusicsegment->SetParam(GUID_StandardMIDIFile, -1, 0, 0, (void*)directmusicperformance);
      directmusicsegment->SetParam(GUID_Download, -1, 0, 0, (void*)directmusicperformance);
      directmusicsegment->SetRepeats(-1);
      //directmusicperformance->PlaySegment(directmusicsegment, 0, 0, &directmusicsegmentstate);
    #else
      char temp[520];
      char ret[520];

      sprintf(temp, "open \"%s\" type sequencer alias midifile", filename);
      if (mciSendString(temp, ret, 512, NULL) != 0) // if it failed
        return false;

      UpdateSystem();
    #endif

    m_MusicType = MUSIC_MIDI;
  }
  else if (stricmp(filename + strlen(filename) - 4, ".s3m") == 0 ||
           stricmp(filename + strlen(filename) - 4, ".mod") == 0 ||
           stricmp(filename + strlen(filename) - 3, ".it") == 0  ||
           stricmp(filename + strlen(filename) - 3, ".xm") == 0)
  {
    module = FMUSIC_LoadSong((char*)filename);
    if (module == NULL)
      return false;

    m_MusicType = MUSIC_MOD;
  }
  else if (stricmp(filename + strlen(filename) - 4, ".wav") == 0)
  {
    stream = FSOUND_Stream_OpenWav((char*)filename, BaseMode | FSOUND_STREAMABLE | FSOUND_LOOP_NORMAL);
    if (stream == NULL)
      return false;

    m_MusicType = MUSIC_WAV;
  }
  else // mp3
  {
    stream = FSOUND_Stream_OpenMpeg((char*)filename, BaseMode | FSOUND_STREAMABLE | FSOUND_LOOP_NORMAL);
    if (stream == NULL)
      return false;

    m_MusicType = MUSIC_MP3;
  }

  m_Loaded = true;
  return true;

#endif
}

////////////////////////////////////////////////////////////////////////////////

void
CMusicModule::Destroy()
{
#ifdef ENABLE_MUSICMODULES

  if (!AudioEnabled)
    return;

  if (!m_Loaded)
    return;

  Stop();

  if (m_MusicType == MUSIC_MIDI)
  {
    #ifdef ENABLE_DIRECTMUSIC
      directmusicperformance->Stop(NULL, NULL, 0, 0);
      directmusicsegmentstate->Release();
      directmusicsegment->Release();
      directmusicperformance->CloseDown();
      directmusicperformance->Release();
      directmusicloader->Release();
      directmusic->Release();
    #else
      char ret[520];
      mciSendString("close midifile", ret, 512, NULL);
      UpdateSystem();
    #endif

    fclose(fpointer);
  }
  else if (m_MusicType == MUSIC_MOD)
  {
    FMUSIC_FreeSong(module);
  }
  else // mp3 & wav
  {
    FSOUND_Stream_Close(stream);
  }

  m_Loaded = false;

#else

  fclose(fpointer);

#endif
}

////////////////////////////////////////////////////////////////////////////////

void
CMusicModule::Play()
{
#ifdef ENABLE_MUSICMODULES

  if (!AudioEnabled)
    return;

  if (m_MusicType == MUSIC_MIDI)
  {
    #ifdef ENABLE_DIRECTMUSIC
      directmusicperformance->PlaySegment(_mm->directmusicsegment, 0, 0, NULL);
    #else
      char ret[512];
      mciSendString("seek midifile to start", ret, 500, SphereWindow);
      mciSendString("play midifile notify", ret, 500, SphereWindow);
      UpdateSystem();
    #endif
  }
  else if (m_MusicType == MUSIC_MOD)
  {
    FMUSIC_PlaySong(module);
    FMUSIC_SetMasterVolume(module, MusicVolume);
  }
  else
  {
    m_StreamChannel = FSOUND_Stream_Play(FSOUND_FREE, stream);
    FSOUND_SetVolumeAbsolute(m_StreamChannel, MusicVolume);
  }

  m_IsPlaying = true;

#endif 
}
////////////////////////////////////////////////////////////////////////////////

void
CMusicModule::Stop()
{
#ifdef ENABLE_MUSICMODULES

  if (!AudioEnabled)
    return;

  if (!m_Loaded)
    return;

  if (m_IsPlaying)
  {
    if (m_MusicType == MUSIC_MIDI)
    {
      #ifdef ENABLE_DIRECTMUSIC
        directmusicperformance->Stop(NULL, NULL, 0, 0);
      #else
        char ret[512];
        mciSendString("stop midifile", ret, 500, NULL);
        mciSendString("seek midifile to start", ret, 500, NULL);
        UpdateSystem();
      #endif
    }
    else if (m_MusicType == MUSIC_MOD)
    {
      FMUSIC_StopSong(module);
    }
    else
    {
      FSOUND_Stream_Stop(stream);
    }
    
    m_IsPlaying = false;
  }

  m_Loaded = false;

#endif
}

////////////////////////////////////////////////////////////////////////////////

unsigned long 
CMusicModule::GetID()
{
  return fhandle;
}

////////////////////////////////////////////////////////////////////////////////

void 
CMusicModule::SetID(unsigned long id)
{
  fhandle = id;
}

////////////////////////////////////////////////////////////////////////////////

FILE* 
CMusicModule::GetFile()
{
  return fpointer;
}

////////////////////////////////////////////////////////////////////////////////

void 
CMusicModule::SetFile(FILE* file)
{
  fpointer = file;
}

////////////////////////////////////////////////////////////////////////////////

long
CMusicModule::GetFileSize()
{
  return m_Size;
}

////////////////////////////////////////////////////////////////////////////////

void
CMusicModule::SetFileSize(long size)
{
  m_Size = size;
}

////////////////////////////////////////////////////////////////////////////////

long
CMusicModule::GetFileStartPosition()
{
  return m_FileStartPosition;
}

////////////////////////////////////////////////////////////////////////////////

void
CMusicModule::SetFileStartPosition(long start)
{
  m_FileStartPosition = start;
}

////////////////////////////////////////////////////////////////////////////////

void
CMusicModule::SetVolume(int volume)
{
#ifdef ENABLE_MUSICMODULES

  if (!AudioEnabled)
    return;

  if (!m_Loaded)
    return;

  if (!m_IsPlaying)
    return;

  if (m_MusicType == MUSIC_MOD)
    FMUSIC_SetMasterVolume(module, MusicVolume);
  else if (m_MusicType == MUSIC_WAV || m_MusicType == MUSIC_MP3)
    FSOUND_SetVolumeAbsolute(m_StreamChannel, MusicVolume);

#endif
}

////////////////////////////////////////////////////////////////////////////////

bool SetGlobalSoundVolume(int volume)
{
  if (AudioEnabled == false)
    return true;

  FSOUND_SetSFXMasterVolume(volume);
  return true;
}

////////////////////////////////////////////////////////////////////////////////

int GetGlobalSoundVolume()
{
  if (AudioEnabled == false)
    return 0;

  return FSOUND_GetSFXMasterVolume();
}

////////////////////////////////////////////////////////////////////////////////

bool SetGlobalMusicVolume(int volume)
{
  if (AudioEnabled == false)
    return true;

  MusicVolume = volume;
  return true;
}

////////////////////////////////////////////////////////////////////////////////

int GetGlobalMusicVolume()
{
  if (AudioEnabled == false)
    return 0;

  return MusicVolume;
}

////////////////////////////////////////////////////////////////////////////////
